@@ -1,7 +1,7 @@
 /*
  * Dynamic Lua binding to GObject using dynamic gobject-introspection.
  *
- * Copyright (c) 2010, 2011, 2012 Pavel Holejsovsky
+ * Copyright (c) 2010, 2011, 2012, 2013 Pavel Holejsovsky
  * Licensed under the MIT license:
  * http://www.opensource.org/licenses/mit-license.php
  *
@@ -64,7 +64,7 @@ static int record_cache;
 static int parent_cache;
 
 gpointer
-lgi_record_new (lua_State *L, int count)
+lgi_record_new (lua_State *L, int count, gboolean alloc)
 {
   Record *record;
   size_t size;
@@ -73,18 +73,27 @@ lgi_record_new (lua_State *L, int count)
 
   /* Calculate size of the record to allocate. */
   lua_getfield (L, -1, "_size");
-  size = G_STRUCT_OFFSET (Record, data) + lua_tonumber (L, -1) * count;
+  size = lua_tonumber (L, -1) * count;
   lua_pop (L, 1);
 
   /* Allocate new userdata for record object, attach proper
      metatable. */
-  record = lua_newuserdata (L, size);
+  record = lua_newuserdata (L, G_STRUCT_OFFSET (Record, data) +
+			    (alloc ? 0 : size));
   lua_pushlightuserdata (L, &record_mt);
   lua_rawget (L, LUA_REGISTRYINDEX);
   lua_setmetatable (L, -2);
-  record->addr = record->data;
-  memset (record->addr, 0, size - G_STRUCT_OFFSET (Record, data));
-  record->store = RECORD_STORE_EMBEDDED;
+  if (G_LIKELY (!alloc))
+    {
+      record->addr = record->data;
+      memset (record->addr, 0, size);
+      record->store = RECORD_STORE_EMBEDDED;
+    }
+  else
+    {
+      record->addr = g_malloc0 (size);
+      record->store = RECORD_STORE_ALLOCATED;
+    }
 
   /* Get ref_repo table, attach it as an environment. */
   lua_pushvalue (L, -2);
@@ -114,6 +123,50 @@ lgi_record_new (lua_State *L, int count)
   return record->addr;
 }
 
+static void
+record_free (lua_State *L, Record *record, int narg)
+{
+  GType gtype;
+  g_assert (record->store == RECORD_STORE_ALLOCATED);
+  lua_getfenv (L, narg);
+  for (;;)
+    {
+      lua_getfield (L, -1, "_gtype");
+      gtype = (GType) lua_touserdata (L, -1);
+      lua_pop (L, 1);
+      if (G_TYPE_IS_BOXED (gtype))
+	{
+	  g_boxed_free (gtype, record->addr);
+	  break;
+	}
+      else
+	{
+	  /* Use custom _free function. */
+	  void (*free_func)(gpointer) =
+	    lgi_gi_load_function (L, -1, "_free");
+	  if (free_func)
+	    {
+	      free_func (record->addr);
+	      break;
+	    }
+	}
+
+      /* Try to get parent of the type. */
+      lua_getfield (L, -1, "_parent");
+      lua_replace (L, -2);
+      if (lua_isnil (L, -1))
+	{
+	  lua_getfenv (L, 1);
+	  lua_getfield (L, -1, "_name");
+	  g_warning ("unable to free record %s, leaking it",
+		     lua_tostring (L, -1));
+	  lua_pop (L, 2);
+	  break;
+	}
+    }
+  lua_pop (L, 1);
+}
+
 void
 lgi_record_2lua (lua_State *L, gpointer addr, gboolean own, int parent)
 {
@@ -152,8 +205,13 @@ lgi_record_2lua (lua_State *L, gpointer addr, gboolean own, int parent)
 	 ownership is properly updated. */
       record = lua_touserdata (L, -1);
       g_assert (record->addr == addr);
-      if (own && record->store == RECORD_STORE_EXTERNAL)
-	record->store = RECORD_STORE_ALLOCATED;
+      if (own)
+	{
+	  if (record->store == RECORD_STORE_EXTERNAL)
+	    record->store = RECORD_STORE_ALLOCATED;
+	  else if (record->store == RECORD_STORE_ALLOCATED)
+	    record_free (L, record, -1);
+	}
 
       return;
     }
@@ -321,7 +379,16 @@ lgi_record_2c (lua_State *L, int narg, gpointer target, gboolean by_value,
 	{
 	  /* Caller wants to steal ownership from us. */
 	  if (G_LIKELY (record->store == RECORD_STORE_ALLOCATED))
-	    record->store = RECORD_STORE_EXTERNAL;
+	    {
+	      /* Check, whether refrepo table specifies custom _refsink
+		 function. */
+	      void (*refsink_func)(gpointer) =
+		lgi_gi_load_function (L, narg, "_refsink");
+	      if (refsink_func)
+		refsink_func(record->addr);
+	      else
+		record->store = RECORD_STORE_EXTERNAL;
+	    }
 	  else
 	    g_critical ("attempt to steal record ownership from unowned rec");
 	}
@@ -367,46 +434,8 @@ record_gc (lua_State *L)
 	uninit (record->addr);
     }
   else if (record->store == RECORD_STORE_ALLOCATED)
-    {
-      /* Free the owned record. */
-      GType gtype;
-      lua_getfenv (L, 1);
-      for (;;)
-	{
-	  lua_getfield (L, -1, "_gtype");
-	  gtype = (GType) lua_touserdata (L, -1);
-	  lua_pop (L, 1);
-	  if (G_TYPE_IS_BOXED (gtype))
-	    {
-	      g_boxed_free (gtype, record->addr);
-	      break;
-	    }
-	  else
-	    {
-	      /* Use custom _free function. */
-	      void (*free_func)(gpointer) =
-		lgi_gi_load_function (L, -1, "_free");
-	      if (free_func)
-		{
-		  free_func (record->addr);
-		  break;
-		}
-	    }
-
-	  /* Try to get parent of the type. */
-	  lua_getfield (L, -1, "_parent");
-	  lua_replace (L, -2);
-	  if (lua_isnil (L, -1))
-	    {
-	      lua_getfenv (L, 1);
-	      lua_getfield (L, -1, "_name");
-	      g_warning ("unable to record_gc %s, leaking it",
-			 lua_tostring (L, -1));
-	      lua_pop (L, 2);
-	      break;
-	    }
-	}
-    }
+    /* Free the owned record. */
+    record_free (L, record, 1);
 
   if (record->store == RECORD_STORE_NESTED)
     {
@@ -423,14 +452,24 @@ static int
 record_tostring (lua_State *L)
 {
   Record *record = record_get (L, 1);
-  lua_pushfstring (L, "lgi.rec %p:", record->addr);
   lua_getfenv (L, 1);
-  lua_getfield (L, -1, "_name");
-  lua_replace (L, -2);
-  if (!lua_isnil (L, -1))
-    lua_concat (L, 2);
+  lua_getfield (L, -1, "_tostring");
+  if (lua_isnil (L, -1))
+    {
+      lua_pop (L, 1);
+      lua_pushfstring (L, "lgi.rec %p:", record->addr);
+      lua_getfield (L, -2, "_name");
+      if (!lua_isnil (L, -1))
+	lua_concat (L, 2);
+      else
+	lua_pop (L, 1);
+    }
   else
-    lua_pop (L, 1);
+    {
+      lua_pushvalue (L, 1);
+      lua_call (L, 1, 1);
+    }
+
   return 1;
 }
 
@@ -480,7 +519,7 @@ static const struct luaL_Reg record_meta_reg[] = {
    unless 'addr' argument (lightuserdata or integer) is specified, in
    which case wraps specified address as record.  Lua prototype:
 
-   recordinstance = core.record.new(repotable[, nil[, count]])
+   recordinstance = core.record.new(repotable[, nil[, count[, alloc]]])
    recordinstance = core.record.new(repotable[, addr[, own]])
 
    own (default false) means whether Lua takes record ownership
@@ -492,9 +531,10 @@ record_new (lua_State *L)
   if (lua_isnoneornil (L, 2))
     {
       /* Create new record instance. */
+      gboolean alloc = lua_toboolean (L, 4);
       luaL_checktype (L, 1, LUA_TTABLE);
       lua_pushvalue (L, 1);
-      lgi_record_new (L, luaL_optinteger (L, 3, 1));
+      lgi_record_new (L, luaL_optinteger (L, 3, 1), alloc);
     }
   else
     {
@@ -548,7 +588,13 @@ record_query (lua_State *L)
   else
     {
       if (lua_isnoneornil (L, 3))
-	lua_pushlightuserdata (L, record_check (L, 1)->addr);
+	{
+	  record = record_check (L, 1);
+	  if (!record)
+	    return 0;
+
+	  lua_pushlightuserdata (L, record->addr);
+	}
       else
 	{
 	  gpointer addr;
@@ -577,6 +623,7 @@ record_field (lua_State *L)
   record = record_get (L, 1);
 
   /* Call field marshalling worker. */
+  lua_getfenv (L, 1);
   return lgi_marshal_field (L, record->addr, getmode, 1, 2, 3);
 }
 
